#ifndef _RHEO_RIESZ_H
#define _RHEO_RIESZ_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================
#include "rheolef/field.h"
#include "rheolef/band.h"
#include "rheolef/test.h"
#include "rheolef/integrate.h"
//#include "rheolef/field_vf_expr.h"
//#include "rheolef/field_vf_expr_ops.h"

#include <boost/type_traits/is_class.hpp>
#include <boost/type_traits/is_function.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/or.hpp>
#include <boost/mpl/not.hpp>

namespace rheolef { 

// --------------------------------------------------------------------------
// utility to determine whether a template arg is a function or a constant
// --------------------------------------------------------------------------
namespace details {
  // build a function that returns a constant
  template<class T1, class T2>
  struct f_constant : field_functor<f_constant<T1,T2>,T2 > {
    const T2& operator() (const T1& x) const { return c; }
    f_constant (const T2& c0) : c(c0) {}
    T2 c;
  };
  template<class F> struct is_vector_function :
    boost::mpl::and_ <
        boost::is_class<F>
       ,is_point<typename F::result_type>
    > {};
  template<class T> struct is_vector_function<point_basic<T> (const point_basic<T>)>
  	: boost::mpl::true_ {};

  template<class E> struct is_expr                    : boost::mpl::false_ {};
  template<class E> struct is_expr<field_nonlinear_expr<E> > : boost::mpl::true_ {};

  template<class C>
  struct is_constant :
        boost::mpl::or_ <
          is_scalar<C>
         ,is_point<C>
        > {};

  template<class F>
  struct is_function :
    boost::mpl::and_ <
      boost::mpl::or_ <
        boost::is_class<F>
       ,boost::is_pointer<F>
       ,boost::is_function<F>
      >
     ,boost::mpl::not_ <
        boost::mpl::or_ <
          is_scalar<F>
         ,is_point<F>
         ,is_expr<F>
        >
      >
    >
  {};
  template<class F>
  struct result_type {
    typedef typename F::result_type type;
  };
  template<class T, class R>
  struct result_type<R(const point_basic<T>&)> {
    typedef R type;
  };
  template<class T, class R>
  struct result_type<R (*)(const point_basic<T>&)> {
    typedef R type;
  };
  template<class Constant>
  struct constant_promote { typedef Constant type; };
  template<>        struct constant_promote<int>    { typedef Float type; };
  template<>        struct constant_promote<size_t> { typedef Float type; };
    
} // namespace details
// --------------------------------------------------------------------------
// generic implementation
// --------------------------------------------------------------------------
// scalar-valued function
template <class T, class M, class Geo, class Function>
field_basic<T,M>
riesz_function_internal (
    const space_basic<T,M>&         Xh,
    const Function&                 f,
    const quadrature_option_type&   qopt,
    const Geo&                      dom,
    boost::mpl::false_)
{
  test_basic<T,M,details::vf_tag_01> v (Xh);
  return integrate (dom, f*v, qopt);
}
// vector-valued function
template <class T, class M, class Geo, class Function>
field_basic<T,M>
riesz_function_internal (
    const space_basic<T,M>&         Xh,
    const Function&                 f,
    const quadrature_option_type&   qopt,
    const Geo&                      dom,
    boost::mpl::true_)
{
  test_basic<T,M,details::vf_tag_01> v (Xh);
  return integrate (dom, dot(f,v), qopt);
}
// --------------------------------------------------------------------------
// switch between compile-time return type:
// --------------------------------------------------------------------------
template <class T, class M, class Geo, class Function>
field_basic<T,M>
riesz_tag2 (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const quadrature_option_type& qopt,
    const Geo&                    dom,
    const T&)
{
  return riesz_function_internal (Xh, f, qopt, dom, mpl::false_());
}
template <class T, class M, class Geo, class Function>
field_basic<T,M>
riesz_tag2 (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const quadrature_option_type& qopt,
    const Geo&                    dom,
    const point_basic<T>&)
{
  return riesz_function_internal (Xh, f, qopt, dom, mpl::true_());
}
template <class T, class M, class Geo, class Function>
field_basic<T,M>
riesz_tag2 (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const quadrature_option_type& qopt,
    const Geo&                    dom,
    const undeterminated_basic<T>&)
{
  if (Xh.valued_tag() == space_constant::scalar) {
    return riesz_function_internal (Xh, f, qopt, dom, mpl::false_());
  } else {
    return riesz_function_internal (Xh, f, qopt, dom, mpl::true_());
  }
}
// --------------------------------------------------------------------------
// f: switch between constant or function
// --------------------------------------------------------------------------
// f is a function:
template <class T, class M, class Geo, class Function>
field_basic<T,M>
riesz_tag (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const quadrature_option_type& qopt,
    const Geo&                    dom,
    boost::mpl::false_)
{
  return riesz_tag2 (Xh, f, qopt, dom, 
	typename details::result_type<Function>::type());
}
// f is a constant:
// thanks to a wrapper class, we go to the first case
template <class T, class M, class Geo, class Constant>
field_basic<T,M>
riesz_tag (
    const space_basic<T,M>&       Xh,
    const Constant&               value,
    const quadrature_option_type& qopt,
    const Geo&                    dom,
    boost::mpl::true_)
{
  typedef typename field_basic<T,M>::float_type float_type;
  typedef typename details::constant_promote<Constant>::type constant_type; // int to Float, but conserve point_basic<T> & tensor
  return riesz_tag (Xh, details::f_constant<point_basic<float_type>,constant_type>(value),
			     qopt, dom, boost::mpl::false_());
}

/*Class:riesz
NAME: @code{riesz} - approximate a Riesz representer
@findex  riesz
@cindex  riesz representer
@cindex  quadrature formula
@clindex space
@clindex field

SYNOPSYS:
@example
 template <class Function>
 field riesz (const space& Xh, const Expr& expr,
    	quadrature_option_type qopt = default_value);

 template <class Function>
 field riesz (const space& Xh, const Expr& expr,
	const geo& domain, quadrature_option_type qopt = default_value);
@end example
NOTE:
  This function has been now suppersetted by the @code{integrate} function
  @pxref{integrate algorithm}. The old-fashioned code:
  @example
     field l1h = riesz (Xh, f);
     field l2h = riesz (Xh, f, "boundary");
  @end example
  writes now:
  @example
     test v (Xh);
     field l1h = integrate (f*v);
     field l2h = integrate ("boundary", f*v);
  @end example
  The @code{riesz} function is still present for backward compatibility purpose.
DESCRIPTION:
  Let @code{f} be any continuous function, its Riesz representer in the finite
  element space @code{Xh} on the domain @code{Omega} is defind by:
@example
                /
                |
  dual(lh,vh) = |      f(x) vh(x) dx
                |
               / Omega
@end example
  for all @code{vh} in @code{Xh}, where @code{dual} denotes the duality
  between @code{Xh} and its dual. As @code{Xh} is a finite dimensional space,
  its dual is identified as @code{Xh} and the duality product as the Euclidian one.
  The Riesz representer is thus the @code{lh} field of @code{Xh} where 
  its i-th degree of freedom is:
@example
                /
                |
  dual(lh,vh) = |      f(x) phi_i(x) dx
                |
               / Omega
@end example
  where phi_i is the i-th basis function in @code{Xh}.
  The integral is evaluated by using a quadrature formula.
  By default the quadrature formule is the Gauss one with
  the order equal to @code{2*k-1} where $@code{k} is the polynomial degree in @code{Xh}.
  Alternative quadrature formula and order is available
  by passing an optional variable to riesz.

@noindent
  The function @code{riesz} implements the
  approximation of the Riesz representer 
  by using some quadrature formula for the evaluation of the integrals.
  Its argument can be any function, class-function or linear or nonlinear
  expressions mixing fields and continuous functions.

@cindex  integrate
@findex  integrate
EXAMPLE:
@noindent
  The following code compute the Riesz representant, denoted
  by @code{lh} of f(x), and the integral of f over the domain omega:
@example
  Float f(const point& x);
  ...
  space Xh (omega_h, "P1");
  field lh = riesz (Xh, f);
  Float int_f = dual(lh, 1);
@end example
OPTIONS:
  An optional argument specifies the quadrature formula used
  for the computation of the integral.
  The domain of integration is by default the mesh associated to
  the finite element space.
  An alternative domain @code{dom}, e.g. a part of 
  the boundary can be supplied as an extra argument.
  This domain can be also a @code{band} associated to
  the banded level set method.
End: */

//<riesz:
template <class T, class M, class Function>
field_basic<T,M>
riesz (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const quadrature_option_type& qopt
       = quadrature_option_type(quadrature_option_type::max_family,0))
//>riesz:
{
  return riesz_tag (Xh, f, qopt, Xh.get_geo(),
	typename details::is_constant<Function>::type());
}
//<riesz:
template <class T, class M, class Function>
field_basic<T,M>
riesz (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const geo_basic<T,M>&         dom,
    const quadrature_option_type& qopt
       = quadrature_option_type(quadrature_option_type::max_family,0))
//>riesz:
{
  return riesz_tag (Xh, f, qopt, dom,
	typename details::is_constant<Function>::type());
}
//<riesz:
template <class T, class M, class Function>
field_basic<T,M>
riesz (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    std::string                   dom_name,
    const quadrature_option_type& qopt
       = quadrature_option_type(quadrature_option_type::max_family,0))
//>riesz:
{
  const geo_basic<T,M>& omega = Xh.get_geo();
  return riesz (Xh, f, omega[dom_name], qopt);
}
//<riesz:
template <class T, class M, class Function>
field_basic<T,M>
riesz (
    const space_basic<T,M>&       Xh,
    const Function&               f,
    const band_basic<T,M>&        gh,
    const quadrature_option_type& qopt
       = quadrature_option_type(quadrature_option_type::max_family,0))
//>riesz:
{
  return riesz_tag (Xh, f, qopt, gh,
	typename details::is_constant<Function>::type());
}

}// namespace rheolef
#endif // _RHEO_RIESZ_H
